from __future__ import annotations

from typing import Literal, Union

from pydantic import Field, RootModel

from arakawa.types import ComputeMethod, SelectType, VAlign

from .asset import Attachment, DataTable, Media, Plot, Table
from .base import DataBlock
from .controls import (
    ChoiceField,
    DateTimeField,
    FileField,
    HiddenField,
    MultiChoiceField,
    NumberBox,
    RangeField,
    SwitchField,
    TagsField,
    TextBox,
)
from .fields import TypeAliasedField
from .misc_blocks import BigNumber, Empty
from .mixins import OptionalLabelMixin, OptionalNameMinx
from .text import HTML, Code, Embed, Text

ControlBlocks = Union[
    ChoiceField,
    DateTimeField,
    FileField,
    HiddenField,
    MultiChoiceField,
    NumberBox,
    RangeField,
    SwitchField,
    TagsField,
    TextBox,
]
AssetBlocks = Union[Media, Attachment, Plot, Table, DataTable]
TextBlocks = Union[Code, Embed, Text, HTML]
MiscBlocks = Union[BigNumber, Empty]
LayoutBlocks = Union["Select", "Group", "Toggle", "Compute"]
AllBlocks = Union[ControlBlocks, AssetBlocks, TextBlocks, MiscBlocks, LayoutBlocks]


class LayoutBlock(OptionalNameMinx, OptionalLabelMixin, DataBlock):
    blocks: list[AllBlocks] = Field(..., min_length=1)


class Page(LayoutBlock):
    type_: Literal["_Page"] = TypeAliasedField()

    title: str | None = Field(default=None, min_length=1, max_length=256)


class Group(LayoutBlock):
    type_: Literal["Group"] = TypeAliasedField()

    columns: int = Field(default=1)
    valign: VAlign
    widths: str | None = Field(default=None, min_length=2, pattern=r"\[\d+(,\s*\d+)*\]")


class Select(LayoutBlock):
    type_: Literal["Select"] = TypeAliasedField()

    type: SelectType


class Toggle(LayoutBlock):
    type_: Literal["Toggle"] = TypeAliasedField()


class Compute(LayoutBlock):
    type_: Literal["Compute"] = TypeAliasedField()

    prompt: str | None = Field(default=None)
    subtitle: str | None = Field(default=None)
    action: str | None = Field(default="")
    method: ComputeMethod = Field(default=ComputeMethod.GET)


class View(LayoutBlock):
    type_: Literal["View"] = TypeAliasedField()

    fragment: bool
    version: int = Field(..., ge=1)


class BlocksWrapper(RootModel):
    root: list[AllBlocks]



"""
# TODO - optimize import handling here
"""
# ruff: noqa: FA100

import json
import pickle
from contextlib import suppress
from io import TextIOWrapper
from typing import Any, Union

import pandas as pd
from altair.utils import SchemaBase
from multimethod import multimethod
from pandas.io.formats.style import Styler

from arakawa import optional_libs as opt
from arakawa.common import ArrowFormat
from arakawa.exceptions import ARError
from arakawa.utils import log

from .pydantic_visitor import AssetMeta


class ARTextIOWrapper(TextIOWrapper):
    """Custom IO Wrapper that detaches before closing - see https://bugs.python.org/issue21363"""

    def __init__(self, f, *a, **kw):
        super().__init__(f, *a, encoding="utf-8", **kw)

    def __del__(self):
        # don't close the underlying stream
        with suppress(Exception):
            self.flush()
        with suppress(Exception):
            self.detach()


class AttachmentWriter:
    @multimethod
    def get_meta(self, _: Any) -> AssetMeta:
        return AssetMeta(ext=".pkl", mime="application/vnd.pickle+binary")

    @get_meta.register  # type: ignore
    def _(self, _: str) -> AssetMeta:
        return AssetMeta(ext=".json", mime="application/json")

    @multimethod
    def write_file(self, x: Any, f) -> None:
        pickle.dump(x, f)

    @write_file.register  # type: ignore
    def _(self, x: str, f) -> None:
        out: str = json.dumps(json.loads(x))
        f.write(out.encode())


class DataTableWriter:
    def get_meta(self, _: pd.DataFrame) -> AssetMeta:
        return AssetMeta(mime=ArrowFormat.content_type, ext=ArrowFormat.ext)

    def write_file(self, x: pd.DataFrame, f) -> None:
        if x.size == 0:
            raise ARError("Empty DataFrame provided")
        # process_df called in Arrow.save_file
        ArrowFormat.save_file(f, x)


class HTMLTableWriter:
    def get_meta(self, _: Union[pd.DataFrame, Styler]) -> AssetMeta:
        return AssetMeta(mime="application/vnd.arakawa.table+html", ext=".tbl.html")

    @multimethod
    def write_file(self, x: pd.DataFrame, f) -> None:
        self._check(x)
        out = x.to_html().encode()
        f.write(out)

    @write_file.register  # type: ignore
    def _(self, x: Styler, f) -> None:
        self._check(x.data)  # type: ignore
        out = x.to_html().encode()
        f.write(out)

    def _check(self, df: pd.DataFrame) -> None:
        n_cells = df.shape[0] * df.shape[1]
        if n_cells > 500:
            log.warning(
                "Table is over recommended size, consider using ar.DataTable instead or aggregating the df first"
            )


class PlotWriter:
    obj_type: Any

    # Altair (always installed)
    @multimethod
    def get_meta(self, _: SchemaBase) -> AssetMeta:
        return AssetMeta(mime="application/vnd.vegalite.v5+json", ext=".vl.json")

    @multimethod
    def write_file(self, x: SchemaBase, f) -> None:
        json.dump(x.to_dict(), ARTextIOWrapper(f))

    if opt.HAVE_FOLIUM:

        @get_meta.register  # type: ignore
        def _(self, _: opt.Map) -> AssetMeta:
            return AssetMeta(mime="application/vnd.folium+html", ext=".fl.html")

        @write_file.register  # type: ignore
        def _(self, x: opt.Map, f) -> None:
            html: str = x.get_root().render()
            f.write(html.encode())

    if opt.HAVE_BOKEH:

        @get_meta.register  # type: ignore
        def _(self, _: Union[opt.BFigure, opt.BLayout]) -> AssetMeta:
            return AssetMeta(mime="application/vnd.bokeh.show+json", ext=".bokeh.json")

        @write_file.register  # type: ignore
        def _(self, x: Union[opt.BFigure, opt.BLayout], f):
            from bokeh.embed import json_item

            json.dump(json_item(x), ARTextIOWrapper(f))

    if opt.HAVE_PLOTLY:

        @get_meta.register  # type: ignore
        def _(self, _: opt.PFigure) -> AssetMeta:
            return AssetMeta(mime="application/vnd.plotly.v1+json", ext=".pl.json")

        @write_file.register  # type: ignore
        def _(self, x: opt.PFigure, f):
            json.dump(x.to_json(), ARTextIOWrapper(f))

    if opt.HAVE_MATPLOTLIB:

        @get_meta.register  # type: ignore
        def _(self, _: Union[opt.Axes, opt.Figure, opt.ndarray]) -> AssetMeta:
            return AssetMeta(mime="image/svg+xml", ext=".svg")

        @write_file.register  # type: ignore
        def _(self, x: opt.Figure, f) -> None:
            x.savefig(ARTextIOWrapper(f), format="svg", bbox_inches="tight")

        @write_file.register  # type: ignore
        def _(self, x: opt.Axes, f) -> None:
            self.write_file(x.get_figure(), f)

        @write_file.register  # type: ignore
        def _(self, x: opt.ndarray, f) -> None:
            fig = x.flatten()[0].get_figure()
            self.write_file(fig, f)



"""Dataset Format handling"""

from __future__ import annotations

import abc
import enum
from typing import IO, Union

import pandas as pd
import pyarrow as pa
from loguru import logger as log
from pandas.errors import ParserError
from pyarrow import RecordBatchFileWriter

from arakawa.types import ARROW_EXT, ARROW_MIMETYPE, MIME

from .df_processor import obj_to_str, process_df, str_to_arrow_str
from .utils import guess_encoding


def write_table(table: pa.Table, sink: str | IO[bytes]):
    """Write an arrow table to a file"""
    writer = RecordBatchFileWriter(sink, table.schema)
    writer.write(table)
    writer.close()


PathOrFile = Union[str, IO]


class DFFormatter(abc.ABC):
    # TODO - tie to mimetypes lib
    content_type: MIME
    ext: str
    enum: str

    @staticmethod
    @abc.abstractmethod
    def load_file(fn: PathOrFile) -> pd.DataFrame:
        pass

    @staticmethod
    @abc.abstractmethod
    def save_file(fn: PathOrFile, df: pd.DataFrame):
        pass


DFFormatterCls = type[DFFormatter]


class ArrowFormat(DFFormatter):
    content_type = ARROW_MIMETYPE
    ext = ARROW_EXT
    enum = "ARROW"

    @staticmethod
    def load_file(fn: PathOrFile) -> pd.DataFrame:
        df = pa.ipc.open_file(fn).read_pandas()
        # NOTE - need to convert categories from object to string https://github.com/apache/arrow/issues/33070
        obj_to_str(df)
        str_to_arrow_str(df)
        return df

    @staticmethod
    def save_file(fn: PathOrFile, df: pd.DataFrame):
        df = process_df(df)
        # NOTE - can pass expected schema and columns for output df here
        table: pa.Table = pa.Table.from_pandas(df, preserve_index=False)
        write_table(table, fn)


class CSVFormat(DFFormatter):
    content_type = MIME("text/csv")
    ext = ".csv"
    enum = "CSV"

    @staticmethod
    def load_file(fn: PathOrFile) -> pd.DataFrame:
        # TODO - fix
        if not isinstance(fn, str):
            raise ValueError("FObj not yet supported")

        try:
            return pd.read_csv(fn, engine="c", sep=",")
        except UnicodeDecodeError:
            encoding = guess_encoding(fn)
            return pd.read_csv(fn, engine="c", sep=",", encoding=encoding)
        except ParserError as e:
            log.warning(f"Error parsing CSV file ({e}), trying python fallback")
            try:
                return pd.read_csv(fn, engine="python", sep=None)
            except UnicodeDecodeError:
                encoding = guess_encoding(fn)
                return pd.read_csv(fn, engine="python", sep=None, encoding=encoding)

    @staticmethod
    def save_file(fn: PathOrFile, df: pd.DataFrame):
        df.to_csv(fn, index=False)


class ExcelFormat(DFFormatter):
    content_type = MIME(
        "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet"
    )
    ext = ".xlsx"
    enum = "EXCEL"

    @staticmethod
    def load_file(fn: PathOrFile) -> pd.DataFrame:
        return pd.read_excel(fn, engine="openpyxl")

    @staticmethod
    def save_file(fn: PathOrFile, df: pd.DataFrame):
        df.to_excel(fn, index=False, engine="openpyxl")


class DatasetFormats(enum.Enum):
    """Used to switch between the different format handlers"""

    CSV = CSVFormat
    EXCEL = ExcelFormat
    ARROW = ArrowFormat


# TODO - make into enums?
df_ext_map: dict[str, DFFormatterCls] = {x.value.ext: x.value for x in DatasetFormats}



from __future__ import annotations

import packaging.version
import semver

from arakawa.exceptions import ARError


def convert2semver(v: str) -> semver.Version:
    """Converts a PyPI version into a semver version"""
    ver = packaging.version.parse(v)

    prerelease: str = ""
    if ver.pre:
        pep440_prerelease_prefix, number = ver.pre

        if pep440_prerelease_prefix == "a":
            prerelease = "alpha"
        elif pep440_prerelease_prefix == "b":
            prerelease = "beta"
        else:
            raise ARError("Unsupported pre-release prefix", pep440_prerelease_prefix)

        if number > 0:
            prerelease += f".{number}"

    return semver.Version(*ver.release, prerelease=prerelease, build=ver.dev)



from __future__ import annotations

from typing import Literal

from pydantic import Field

from arakawa.types import NumberValue

from .base import DataBlock
from .fields import TypeAliasedField
from .mixins import NameMixin


class Empty(NameMixin, DataBlock):
    type_: Literal["Empty"] = TypeAliasedField()


class BigNumber(DataBlock):
    type_: Literal["BigNumber"] = TypeAliasedField()

    heading: str = Field(..., min_length=1, max_length=128)
    value: NumberValue = Field(..., min_length=1, max_length=128)
    change: NumberValue | None = Field(default=None, min_length=1, max_length=128)
    pre_value: NumberValue | None = Field(default=None, min_length=1, max_length=128)
    is_positive_intent: bool | None = Field(default=None)
    is_upward_change: bool | None = Field(default=None)



from __future__ import annotations

import sys
from abc import ABC
from typing import TYPE_CHECKING, Any, ClassVar, TypeVar, Union, cast

from arakawa.common.utils import is_valid_id, mk_attribs
from arakawa.exceptions import ARError

if sys.version_info <= (3, 11):
    pass
else:
    pass

if TYPE_CHECKING:
    from arakawa.blocks import Block
    from arakawa.view import ViewVisitor


BlockId = str

VV = TypeVar("VV", bound="ViewVisitor")


class BaseBlock(ABC):
    """Base Block class - subclassed by all Block types

    ..note:: The class is not used directly.
    """

    _tag: ClassVar[str]

    def __init__(self, name: BlockId | None = None, **kwargs: Any):
        """
        Args:
            name: A unique name to reference the block, used when referencing blocks via the report editor and when embedding
        """
        self._type = self._tag
        self.name = name

        # validate name
        if name and not is_valid_id(name):
            raise ARError(f"Invalid name '{name}' for block")

        self._add_attributes(**kwargs)

    def _add_attributes(self, **kwargs):
        attrs = mk_attribs(**kwargs)
        for key, value in attrs.items():
            setattr(self, key, value)

    def _ipython_display_(self):
        """Display the block as a side effect within a Jupyter notebook"""
        from IPython.display import HTML, display

        from arakawa.ipython.environment import get_environment
        from arakawa.processors.api import stringify_report
        from arakawa.view import Blocks

        if get_environment().support_rich_display:
            html_str = stringify_report(Blocks(self))
            display(HTML(html_str))
        else:
            display(self.__str__())

    def _accept(self, visitor: VV) -> VV:
        # NOTE: use underscore prefix to avoid conflict
        visitor.visit(self)
        return visitor

    def __copy__(self):
        inst = self.__class__.__new__(self.__class__)
        inst.__dict__.update(self.__dict__)
        return inst


class DataBlock(BaseBlock):
    """Abstract block that represents a leaf-node in the tree, e.g. a Plot or Table

    ..note:: This class is not used directly.
    """


BlockOrPrimitive = Union[BaseBlock, Any]  # TODO - expand
BlockList = list[BaseBlock]


def wrap_block(b: BlockOrPrimitive) -> Block:
    from .wrappers import convert_to_block

    # if isinstance(b, Page):
    #     raise DPClientError("Page objects can only be at the top-level")
    if not isinstance(b, BaseBlock):
        # import here as a very slow module due to nested imports
        # from ..files import convert
        return convert_to_block(b)

    return cast("Block", b)



from .asset import Attachment, DataTable, Media, Plot, Table  # noqa: F401
from .base import (  # noqa: F401
    BaseBlock,
    BlockList,
    BlockOrPrimitive,
    DataBlock,
    wrap_block,
)
from .controls import (  # noqa: F401
    ChoiceField,
    ColorField,
    DateField,
    DateTimeField,
    EmailField,
    FileField,
    HiddenField,
    MultiChoiceField,
    NumberBox,
    PasswordField,
    RangeField,
    SearchField,
    SwitchField,
    TagsField,
    TelephoneField,
    TextareaField,
    TextBox,
    TimeField,
    URLField,
)
from .layout import (  # noqa: F401
    Compute,
    Group,
    Page,
    Select,
    SelectType,
    Toggle,
    VAlign,
)
from .misc_blocks import BigNumber, Empty  # noqa: F401
from .text import HTML, Code, Embed, Formula, Text  # noqa: F401

Block = BaseBlock



from __future__ import annotations

from typing import Literal

from pydantic import AnyHttpUrl, Field

from .base import DataBlock
from .fields import TypeAliasedField
from .mixins import OptionalCaptionMixin, OptionalLabelMixin, OptionalNameMinx


class TextBase(OptionalLabelMixin, OptionalNameMinx, DataBlock):
    content: str = Field(..., min_length=1, pattern=r"(.|\s)*\S(.|\s)*")


class Text(TextBase):
    type_: Literal["Text", "Formula"] = TypeAliasedField()


class HTML(TextBase):
    type_: Literal["HTML"] = TypeAliasedField()

    sandbox: str | None = Field(default=None)


class Code(OptionalCaptionMixin, TextBase):
    type_: Literal["Code"] = TypeAliasedField()

    language: str = Field(..., min_length=1, max_length=128)


class Embed(TextBase):
    type_: Literal["Embed"] = TypeAliasedField()

    url: AnyHttpUrl
    title: str = Field(..., min_length=1, max_length=256)
    provider_name: str = Field(..., min_length=1, max_length=128)



from __future__ import annotations

import importlib.resources as ir
import json
from abc import ABC
from copy import copy
from os import path as osp
from pathlib import Path
from typing import Any, cast
from uuid import uuid4

import humps
from jinja2 import Environment, FileSystemLoader, Template, pass_context
from jinja2.utils import htmlsafe_json_dumps
from markupsafe import Markup

from arakawa import blocks as b
from arakawa.common import timestamp
from arakawa.exceptions import InvalidReportError
from arakawa.types import HTML, NPath
from arakawa.utils import display_msg, open_in_browser
from arakawa.view import PreProcess

from .types import BaseProcessor, Formatting


@pass_context
def include_raw(ctx, name) -> Markup:
    """Normal jinja2 {% include %} doesn't escape {{...}} which appear in React's source code"""
    env = ctx.environment
    # Escape </script> to prevent 3rd party JS terminating the local app bundle.
    # Note there's an extra "\" because it needs to be escaped at both the python and JS level
    src = env.loader.get_source(env, name)[0].replace("</script>", r"<\\/script>")
    return Markup(src)


class PreProcessView(BaseProcessor):
    """Optimization to improve the layout of the view using the Block-API"""

    def __init__(self, *, is_finalized: bool = True) -> None:
        self.is_finalized = is_finalized
        super().__init__()

    def __call__(self, _: Any) -> None:
        # AST checks
        if len(self.s.blocks.blocks) == 0:
            raise InvalidReportError(
                "Empty blocks object - must contain at least one block"
            )

        # convert Page -> Select + Group
        v = copy(self.s.blocks)
        if all(isinstance(blk, b.Page) for blk in v.blocks):
            # convert to top-level Select
            v.blocks = [
                b.Select(
                    blocks=[
                        b.Group(blocks=p.blocks, label=p.title, name=p.name)
                        for p in cast(list[b.Page], v.blocks)
                    ],
                    type=b.SelectType.TABS,
                )
            ]

        # Block-API visitors
        pp = PreProcess(is_finalized=self.is_finalized)
        v._accept(pp)
        v1 = pp.root
        # v1 = copy(v)

        # update the processor state
        self.s.blocks = v1

        return


class ConvertPydantic(BaseProcessor):
    def __init__(self, *, pretty_print: bool = False, fragment: bool = False) -> None:
        self.pretty_print: bool = pretty_print
        self.fragment: bool = fragment
        super().__init__()

    def __call__(self, _: Any):
        from arakawa.view.pydantic_visitor import PydanticBuilder

        builder_state = PydanticBuilder(store=self.s.store)
        self.s.blocks._accept(builder_state)
        view = builder_state.get_root(self.fragment)
        self.s.view_json = humps.camelize(
            view.model_dump(mode="json", by_alias=True, exclude_none=True)
        )
        return view


###############################################################################
# HTML Exporting Processors
class BaseExportHTML(BaseProcessor, ABC):
    """Provides shared logic for writing an app to local disk"""

    # Type is `ir.abc.Traversable` which extends `Path`,
    # but the former isn't compatible with `shutil`
    template_dir: Path = cast(Path, ir.files("arakawa.resources.html_templates"))
    template: Template

    template_name: str

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)

        template_loader = FileSystemLoader(cls.template_dir)
        template_env = Environment(loader=template_loader)
        template_env.globals["include_raw"] = include_raw
        cls.template = template_env.get_template(cls.template_name)

    def get_cdn(self) -> str:
        from arakawa import settings

        return settings.AR_CDN_BASE

    def _write_html_template(
        self,
        name: str,
        formatting: Formatting | None = None,
        cdn_base: str | None = None,
        standalone: bool = False,
    ) -> tuple[str, str]:
        """Internal method to write the ViewXML and assets into a HTML container and associated files"""
        name = name or "app"
        formatting = formatting or Formatting()

        report_id: str = uuid4().hex

        # TODO - split this out?
        vs = self.s
        if vs:
            assets = vs.store.as_dict() or {}
            view_json = vs.view_json
        else:
            assets = {}
            view_json = {}

        app_data = {"viewJson": view_json, "assets": assets}
        html = self.template.render(
            # Escape JS multi-line strings
            app_data=htmlsafe_json_dumps({"data": {"result": app_data}}),
            report_width_class=formatting.width.to_css(),
            report_name=name,
            report_date=timestamp(),
            css_header=formatting.to_css(),
            is_light_prose=json.dumps(formatting.light_prose),
            events=True,
            report_id=report_id,
            cdn_base=cdn_base or self.get_cdn(),
            standalone=standalone,
        )

        return html, report_id


class ExportHTMLInlineAssets(BaseExportHTML):
    """
    Export a view into a single HTML file containing:
    - View XML - embedded
    - Assets - embedded as b64 data-uris
    """

    template_name = "local_template.html.j2"

    def __init__(
        self,
        path: str,
        open: bool = False,
        name: str = "app",
        formatting: Formatting | None = None,
        cdn_base: str | None = None,
        standalone: bool = False,
    ):
        self.path = path
        self.open = open
        self.name = name
        self.formatting = formatting
        self.cdn_base = cdn_base
        self.standalone = standalone

    def __call__(self, _: Any) -> str:
        html, report_id = self._write_html_template(
            name=self.name,
            formatting=self.formatting,
            cdn_base=self.cdn_base,
            standalone=self.standalone,
        )

        Path(self.path).write_text(html, encoding="utf-8")

        display_msg(f"App saved to ./{self.path}")

        if self.open:
            path_uri = f"file://{osp.realpath(osp.expanduser(self.path))}"
            open_in_browser(path_uri)

        return report_id


class ExportHTMLFileAssets(BaseExportHTML):
    """
    Export a view into a single HTML file on disk, containing
    - View XML - embedded
    - Assets - referenced as remote resources
    """

    template_name = "local_template.html.j2"

    def __init__(
        self,
        app_dir: Path,
        name: str = "app",
        formatting: Formatting | None = None,
        cdn_base: str | None = None,
        standalone: bool = False,
    ):
        self.app_dir = app_dir
        self.name = name
        self.formatting = formatting
        self.cdn_base = cdn_base
        self.standalone = standalone

    def __call__(self, _: NPath | None = None) -> Path:
        html, _ = self._write_html_template(
            name=self.name,
            formatting=self.formatting,
            cdn_base=self.cdn_base,
            standalone=self.standalone,
        )

        index_path = self.app_dir / "index.html"
        index_path.write_text(html, encoding="utf-8")
        display_msg(f"Built app in {self.app_dir}")
        return self.app_dir


class ExportHTMLStringInlineAssets(BaseExportHTML):
    """
    Export the View as an in-memory string representing a resizable HTML fragment, containing
    - View XML - embedded
    - Assets - embedded as b64 data-uris
    """

    template_name = "ipython_template.html.j2"

    def __init__(
        self,
        name: str = "Stringified App",
        formatting: Formatting | None = None,
        cdn_base: str | None = None,
        standalone: bool = False,
    ):
        self.name = name
        self.formatting = formatting
        self.cdn_base = cdn_base
        self.standalone = standalone

    def __call__(self, _: Any) -> HTML:
        html, _ = self._write_html_template(
            name=self.name,
            formatting=self.formatting,
            cdn_base=self.cdn_base,
            standalone=self.standalone,
        )
        return HTML(html)


class ExportHTMLStringInlineNonResizableAssets(ExportHTMLStringInlineAssets):
    """
    Export the View as an in-memory string representing a non-resizable HTML fragment, containing
    - View XML - embedded
    - Assets - embedded as b64 data-uris
    """

    template_name = "local_template.html.j2"



from __future__ import annotations

from typing import Literal

from pydantic import Field

from .base import DataBlock
from .fields import TypeAliasedField
from .mixins import NameMixin


class ControlBlock(NameMixin, DataBlock):
    pass


class OptionalHelpMixin(DataBlock):
    help: str | None = Field(default=None)


class OptionalRequiredMixin(DataBlock):
    required: bool | None = Field(default=None)


class OptionalValidationMixin(DataBlock):
    validation: str | None = Field(default=None)


class OptionalStringInitialMixin(DataBlock):
    initial: str | None = Field(default=None)


class DateTimeField(
    OptionalStringInitialMixin,
    OptionalRequiredMixin,
    OptionalValidationMixin,
    OptionalHelpMixin,
    ControlBlock,
):
    type_: Literal["DateTimeField", "DateField", "TimeField"] = Field(
        ..., alias="_type"
    )


class FileField(
    OptionalStringInitialMixin,
    OptionalRequiredMixin,
    OptionalValidationMixin,
    OptionalHelpMixin,
    ControlBlock,
):
    type_: Literal["FileField"] = TypeAliasedField()

    accept: str | None = Field(default=None)


class MultiChoiceField(
    OptionalRequiredMixin,
    OptionalValidationMixin,
    OptionalHelpMixin,
    ControlBlock,
):
    type_: Literal["MultiChoiceField"] = TypeAliasedField()

    initial: list[str] = Field(..., min_length=1)
    options: list[str] = Field(..., min_length=1)


class NumberBox(
    OptionalValidationMixin,
    OptionalHelpMixin,
    OptionalRequiredMixin,
    ControlBlock,
):
    type_: Literal["NumberBox"] = TypeAliasedField()

    initial: int | float | None = Field(default=None)


class RangeField(
    OptionalValidationMixin,
    OptionalHelpMixin,
    OptionalRequiredMixin,
    ControlBlock,
):
    type_: Literal["RangeField"] = TypeAliasedField()

    initial: int | float
    min: int | float
    max: int | float
    step: int | float


class ChoiceField(
    OptionalValidationMixin,
    OptionalStringInitialMixin,
    OptionalHelpMixin,
    OptionalRequiredMixin,
    ControlBlock,
):
    type_: Literal["ChoiceField"] = TypeAliasedField()

    options: list[str] = Field(..., min_length=1)


class SwitchField(
    OptionalValidationMixin,
    OptionalHelpMixin,
    OptionalRequiredMixin,
    ControlBlock,
):
    type_: Literal["SwitchField"] = TypeAliasedField()

    initial: bool | None = Field(default=None)


class TagsField(
    OptionalValidationMixin,
    OptionalHelpMixin,
    OptionalRequiredMixin,
    ControlBlock,
):
    type_: Literal["TagsField"] = TypeAliasedField()

    initial: list[str] = Field(..., min_length=1)


class TextBox(
    OptionalValidationMixin,
    OptionalStringInitialMixin,
    OptionalHelpMixin,
    OptionalRequiredMixin,
    ControlBlock,
):
    type_: Literal[
        "TextBox",
        "TextAreaField",
        "URLField",
        "EmailField",
        "SearchField",
        "TelephoneField",
        "PasswordField",
        "ColorField",
    ] = TypeAliasedField()


class HiddenField(ControlBlock):
    type_: Literal["HiddenField"] = TypeAliasedField()

    initial: str
